@@ -32,7 +32,7 @@ from twisted.internet.protocol import Factory
 from twisted.protocols.basic import Int32StringReceiver
 from twisted.internet.error import ConnectionDone
 
-from autobahn.twisted.util import peer2str
+from autobahn.twisted.util import peer2str, transport_channel_id
 from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost
 
 import txaio
@@ -246,6 +246,12 @@ class WampRawSocketServerProtocol(WampRawSocketProtocol):
             if data:
                 self.dataReceived(data)
 
+    def get_channel_id(self, channel_id_type=u'tls-unique'):
+        """
+        Implements :func:`autobahn.wamp.interfaces.ITransport.get_channel_id`
+        """
+        return transport_channel_id(self.transport, is_server=True, channel_id_type=channel_id_type)
+
 
 class WampRawSocketClientProtocol(WampRawSocketProtocol):
     """
@@ -313,6 +319,12 @@ class WampRawSocketClientProtocol(WampRawSocketProtocol):
             if data:
                 self.dataReceived(data)
 
+    def get_channel_id(self, channel_id_type=u'tls-unique'):
+        """
+        Implements :func:`autobahn.wamp.interfaces.ITransport.get_channel_id`
+        """
+        return transport_channel_id(self.transport, is_server=False, channel_id_type=channel_id_type)
+
 
 class WampRawSocketFactory(Factory):
     """
@@ -345,6 +357,14 @@ class WampRawSocketServerFactory(WampRawSocketFactory):
         if serializers is None:
             serializers = []
 
+            # try CBOR WAMP serializer
+            try:
+                from autobahn.wamp.serializer import CBORSerializer
+                serializers.append(CBORSerializer(batched=True))
+                serializers.append(CBORSerializer())
+            except ImportError:
+                pass
+
             # try MsgPack WAMP serializer
             try:
                 from autobahn.wamp.serializer import MsgPackSerializer
@@ -391,6 +411,15 @@ class WampRawSocketClientFactory(WampRawSocketFactory):
 
         self.debug = debug
 
+        if serializer is None:
+
+            # try CBOR WAMP serializer
+            try:
+                from autobahn.wamp.serializer import CBORSerializer
+                serializer = CBORSerializer()
+            except ImportError:
+                pass
+
         if serializer is None:
 
             # try MsgPack WAMP serializer
