@@ -39,6 +39,7 @@ from autobahn.wamp import exception
 from autobahn.wamp.exception import ApplicationError, ProtocolError, SessionNotReady, SerializationError
 from autobahn.wamp.interfaces import IApplicationSession  # noqa
 from autobahn.wamp.types import SessionDetails
+from autobahn.wamp.keyring import EncryptedPayload
 from autobahn.util import IdGenerator, ObservableMixin
 
 from autobahn.wamp.request import \
@@ -100,6 +101,9 @@ class BaseSession(ObservableMixin):
         self._authmethod = None
         self._authprovider = None
 
+        # end-to-end encryption keyring
+        self._keyring = None
+
         # generator for WAMP request IDs
         self._request_id_gen = IdGenerator()
 
@@ -116,7 +120,7 @@ class BaseSession(ObservableMixin):
             self._ecls_to_uri_pat[exception] = [uri.Pattern(six.u(error), uri.Pattern.URI_TARGET_HANDLER)]
             self._uri_to_ecls[six.u(error)] = exception
 
-    def _message_from_exception(self, request_type, request, exc, tb=None):
+    def _message_from_exception(self, request_type, request, exc, tb=None, enc_algo=None):
         """
         Create a WAMP error message from an exception.
 
@@ -129,6 +133,8 @@ class BaseSession(ObservableMixin):
         :param tb: Optional traceback. If present, it'll be included with the WAMP error message.
         :type tb: list or None
         """
+        assert(enc_algo is None or enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX)
+
         args = None
         if hasattr(exc, 'args'):
             args = list(exc.args)  # make sure tuples are made into lists
@@ -151,7 +157,24 @@ class BaseSession(ObservableMixin):
             else:
                 error = u"wamp.error.runtime_error"
 
-        msg = message.Error(request_type, request, error, args, kwargs)
+        encrypted_payload = None
+        if self._keyring:
+            encrypted_payload = self._keyring.encrypt(False, error, args, kwargs)
+
+        if encrypted_payload:
+            msg = message.Error(request_type,
+                                request,
+                                error,
+                                payload=encrypted_payload.payload,
+                                enc_algo=encrypted_payload.algo,
+                                enc_key=encrypted_payload.pkey,
+                                enc_serializer=encrypted_payload.serializer)
+        else:
+            msg = message.Error(request_type,
+                                request,
+                                error,
+                                args,
+                                kwargs)
 
         return msg
 
@@ -168,6 +191,30 @@ class BaseSession(ObservableMixin):
         # 2. extract additional args/kwargs from error URI
 
         exc = None
+        enc_err = None
+
+        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+
+            if not self._keyring:
+                log_msg = u"received encrypted payload, but no keyring active"
+                self.log.warn(log_msg)
+                enc_err = ApplicationError(ApplicationError.ENC_NO_KEYRING_ACTIVE, log_msg, enc_algo=msg.enc_algo)
+            else:
+                try:
+                    encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
+                    decrypted_error, msg.args, msg.kwargs = self._keyring.decrypt(True, msg.error, encrypted_payload)
+                except Exception as e:
+                    log_msg = u"failed to decrypt application payload 1: {}".format(e)
+                    self.log.warn(log_msg)
+                    enc_err = ApplicationError(ApplicationError.ENC_DECRYPT_ERROR, log_msg, enc_algo=msg.enc_algo)
+                else:
+                    if msg.error != decrypted_error:
+                        log_msg = u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_error, msg.error)
+                        self.log.warn(log_msg)
+                        enc_err = ApplicationError(ApplicationError.ENC_TRUSTED_URI_MISMATCH, log_msg, enc_algo=msg.enc_algo)
+
+        if enc_err:
+            return enc_err
 
         if msg.error in self._uri_to_ecls:
             ecls = self._uri_to_ecls[msg.error]
@@ -207,6 +254,9 @@ class BaseSession(ObservableMixin):
                 else:
                     exc = exception.ApplicationError(msg.error)
 
+        if hasattr(exc, 'enc_algo'):
+            exc.enc_algo = msg.enc_algo
+
         return exc
 
 
@@ -248,6 +298,11 @@ class ApplicationSession(BaseSession):
         # incoming invocations
         self._invocations = {}
 
+    def set_keyring(self, keyring):
+        """
+        """
+        self._keyring = keyring
+
     def onOpen(self, transport):
         """
         Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onOpen`
@@ -265,22 +320,30 @@ class ApplicationSession(BaseSession):
         """
         self.join(self.config.realm)
 
-    def join(self, realm, authmethods=None, authid=None):
+    def join(self, realm, authmethods=None, authid=None, authrole=None, authextra=None):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.join`
         """
+        # FIXME
         if six.PY2 and type(realm) == str:
             realm = six.u(realm)
         if six.PY2 and type(authid) == str:
             authid = six.u(authid)
+        if six.PY2 and type(authrole) == str:
+            authrole = six.u(authrole)
 
         if self._session_id:
             raise Exception("already joined")
 
+        # store the realm requested by client, though this might be overwritten later,
+        # when realm redirection kicks in
+        self._realm = realm
+
+        # closing handshake state
         self._goodbye_sent = False
 
-        msg = message.Hello(realm, role.DEFAULT_CLIENT_ROLES, authmethods, authid)
-        self._realm = realm
+        # send HELLO message to router
+        msg = message.Hello(realm, role.DEFAULT_CLIENT_ROLES, authmethods, authid, authrole, authextra)
         self._transport.send(msg)
 
     def disconnect(self):
@@ -296,6 +359,12 @@ class ApplicationSession(BaseSession):
         """
         return self._transport is not None
 
+    def is_attached(self):
+        """
+        Implements :func:`autobahn.wamp.interfaces.ISession.is_attached`
+        """
+        return self._session_id is not None
+
     def onUserError(self, fail, msg):
         """
         This is called when we try to fire a callback, but get an
@@ -347,9 +416,13 @@ class ApplicationSession(BaseSession):
 
             # the first message must be WELCOME, ABORT or CHALLENGE ..
             if isinstance(msg, message.Welcome):
+
+                if msg.realm:
+                    self._realm = msg.realm
+
                 self._session_id = msg.session
 
-                details = SessionDetails(self._realm, self._session_id, msg.authid, msg.authrole, msg.authmethod)
+                details = SessionDetails(self._realm, self._session_id, msg.authid, msg.authrole, msg.authmethod, msg.authprovider, msg.authextra)
                 d = txaio.as_future(self.onJoin, details)
 
                 def _error(e):
@@ -372,8 +445,12 @@ class ApplicationSession(BaseSession):
                 d = txaio.as_future(self.onChallenge, challenge)
 
                 def success(signature):
-                    if not isinstance(signature, six.text_type):
-                        raise Exception('signature must be unicode')
+                    if signature is None:
+                        raise Exception('onChallenge user callback did not return a signature')
+                    if type(signature) == six.binary_type:
+                        signature = signature.decode('utf8')
+                    if type(signature) != six.text_type:
+                        raise Exception('signature must be unicode (was {})'.format(type(signature)))
                     reply = message.Authenticate(signature)
                     self._transport.send(reply)
 
@@ -424,14 +501,29 @@ class ApplicationSession(BaseSession):
                     for subscription in self._subscriptions[msg.subscription]:
 
                         handler = subscription.handler
+                        topic = msg.topic or subscription.topic
+
+                        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+                            # FIXME: behavior in error cases (no keyring, decrypt issues, URI mismatch, ..)
+                            if not self._keyring:
+                                self.log.warn("received encrypted payload, but no keyring active - ignoring encrypted payload!")
+                            else:
+                                try:
+                                    encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
+                                    decrypted_topic, msg.args, msg.kwargs = self._keyring.decrypt(False, topic, encrypted_payload)
+                                except Exception as e:
+                                    self.log.warn("failed to decrypt application payload: {error}", error=e)
+                                else:
+                                    if topic != decrypted_topic:
+                                        self.log.warn("envelope topic URI does not match encrypted one")
 
                         invoke_args = (handler.obj,) if handler.obj else tuple()
                         if msg.args:
                             invoke_args = invoke_args + tuple(msg.args)
-
                         invoke_kwargs = msg.kwargs if msg.kwargs else dict()
+
                         if handler.details_arg:
-                            invoke_kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher, topic=msg.topic or subscription.topic)
+                            invoke_kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher, topic=topic, enc_algo=msg.enc_algo)
 
                         def _error(e):
                             errmsg = 'While firing {0} subscribed under {1}.'.format(
@@ -452,7 +544,7 @@ class ApplicationSession(BaseSession):
                     publish_request = self._publish_reqs.pop(msg.request)
 
                     # create a new publication object
-                    publication = Publication(msg.publication)
+                    publication = Publication(msg.publication, was_encrypted=publish_request.was_encrypted)
 
                     # resolve deferred/future for publishing successfully
                     txaio.resolve(publish_request.on_reply, publication)
@@ -502,51 +594,76 @@ class ApplicationSession(BaseSession):
 
                 if msg.request in self._call_reqs:
 
-                    if msg.progress:
+                    call_request = self._call_reqs[msg.request]
+                    proc = call_request.procedure
+                    enc_err = None
 
-                        # progressive result
-                        call_request = self._call_reqs[msg.request]
-                        if call_request.options.on_progress:
-                            kw = msg.kwargs or dict()
-                            args = msg.args or tuple()
-                            try:
-                                # XXX what if on_progress returns a Deferred/Future?
-                                call_request.options.on_progress(*args, **kw)
-                            except Exception:
-                                try:
-                                    self.onUserError(
-                                        txaio.create_failure(),
-                                        "While firing on_progress",
-                                    )
-                                except:
-                                    pass
+                    if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
 
+                        if not self._keyring:
+                            log_msg = u"received encrypted payload, but no keyring active"
+                            self.log.warn(log_msg)
+                            enc_err = ApplicationError(ApplicationError.ENC_NO_KEYRING_ACTIVE, log_msg)
                         else:
-                            # silently ignore progressive results
-                            pass
+                            try:
+                                encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
+                                decrypted_proc, msg.args, msg.kwargs = self._keyring.decrypt(True, proc, encrypted_payload)
+                            except Exception as e:
+                                log_msg = u"failed to decrypt application payload 1: {}".format(e)
+                                self.log.warn(log_msg)
+                                enc_err = ApplicationError(ApplicationError.ENC_DECRYPT_ERROR, log_msg)
+                            else:
+                                if proc != decrypted_proc:
+                                    log_msg = u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc)
+                                    self.log.warn(log_msg)
+                                    enc_err = ApplicationError(ApplicationError.ENC_TRUSTED_URI_MISMATCH, log_msg)
+
+                    if msg.progress:
+                        # process progressive call result
+
+                        if call_request.options.on_progress:
+                            if enc_err:
+                                self.onUserError(enc_err, "could not deliver progressive call result, because payload decryption failed")
+                            else:
+                                kw = msg.kwargs or dict()
+                                args = msg.args or tuple()
+                                try:
+                                    # XXX what if on_progress returns a Deferred/Future?
+                                    call_request.options.on_progress(*args, **kw)
+                                except Exception:
+                                    try:
+                                        self.onUserError(txaio.create_failure(), "While firing on_progress")
+                                    except:
+                                        pass
 
                     else:
-                        # final result
-                        #
-                        call_request = self._call_reqs.pop(msg.request)
+                        # process final call result
 
+                        # drop original request
+                        del self._call_reqs[msg.request]
+
+                        # user callback that gets fired
                         on_reply = call_request.on_reply
 
-                        if msg.kwargs:
-                            if msg.args:
-                                res = types.CallResult(*msg.args, **msg.kwargs)
-                            else:
-                                res = types.CallResult(**msg.kwargs)
-                            txaio.resolve(on_reply, res)
+                        # above might already have rejected, so we guard ..
+                        if enc_err:
+                            txaio.reject(on_reply, enc_err)
                         else:
-                            if msg.args:
-                                if len(msg.args) > 1:
-                                    res = types.CallResult(*msg.args)
-                                    txaio.resolve(on_reply, res)
+                            if msg.kwargs:
+                                if msg.args:
+                                    res = types.CallResult(*msg.args, **msg.kwargs)
                                 else:
-                                    txaio.resolve(on_reply, msg.args[0])
+                                    res = types.CallResult(**msg.kwargs)
+                                txaio.resolve(on_reply, res)
                             else:
-                                txaio.resolve(on_reply, None)
+                                if msg.args:
+                                    if len(msg.args) > 1:
+                                        res = types.CallResult(*msg.args)
+                                        txaio.resolve(on_reply, res)
+                                    else:
+                                        txaio.resolve(on_reply, msg.args[0])
+                                else:
+                                    txaio.resolve(on_reply, None)
                 else:
                     raise ProtocolError("RESULT received for non-pending request ID {0}".format(msg.request))
 
@@ -565,78 +682,154 @@ class ApplicationSession(BaseSession):
                     else:
                         registration = self._registrations[msg.registration]
                         endpoint = registration.endpoint
+                        proc = msg.procedure or registration.procedure
+                        enc_err = None
+
+                        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+                            if not self._keyring:
+                                log_msg = u"received encrypted INVOCATION payload, but no keyring active"
+                                self.log.warn(log_msg)
+                                enc_err = ApplicationError(ApplicationError.ENC_NO_KEYRING_ACTIVE, log_msg)
+                            else:
+                                try:
+                                    encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
+                                    decrypted_proc, msg.args, msg.kwargs = self._keyring.decrypt(False, proc, encrypted_payload)
+                                except Exception as e:
+                                    log_msg = u"failed to decrypt INVOCATION payload: {}".format(e)
+                                    self.log.warn(log_msg)
+                                    enc_err = ApplicationError(ApplicationError.ENC_DECRYPT_ERROR, log_msg)
+                                else:
+                                    if proc != decrypted_proc:
+                                        log_msg = u"URI within encrypted INVOCATION payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc)
+                                        self.log.warn(log_msg)
+                                        enc_err = ApplicationError(ApplicationError.ENC_TRUSTED_URI_MISMATCH, log_msg)
+
+                        if enc_err:
+                            # when there was a problem decrypting the INVOCATION payload, we obviously can't invoke
+                            # the endpoint, but return and
+                            reply = self._message_from_exception(message.Invocation.MESSAGE_TYPE, msg.request, enc_err)
+                            self._transport.send(reply)
 
-                        if endpoint.obj is not None:
-                            invoke_args = (endpoint.obj,)
                         else:
-                            invoke_args = tuple()
 
-                        if msg.args:
-                            invoke_args = invoke_args + tuple(msg.args)
-
-                        invoke_kwargs = msg.kwargs if msg.kwargs else dict()
-
-                        if endpoint.details_arg:
-
-                            if msg.receive_progress:
-                                def progress(*args, **kwargs):
-                                    progress_msg = message.Yield(msg.request, args=args, kwargs=kwargs, progress=True)
-                                    self._transport.send(progress_msg)
+                            if endpoint.obj is not None:
+                                invoke_args = (endpoint.obj,)
                             else:
-                                progress = None
+                                invoke_args = tuple()
 
-                            invoke_kwargs[endpoint.details_arg] = types.CallDetails(progress, caller=msg.caller, procedure=msg.procedure)
+                            if msg.args:
+                                invoke_args = invoke_args + tuple(msg.args)
 
-                        on_reply = txaio.as_future(endpoint.fn, *invoke_args, **invoke_kwargs)
+                            invoke_kwargs = msg.kwargs if msg.kwargs else dict()
 
-                        def success(res):
-                            del self._invocations[msg.request]
+                            if endpoint.details_arg:
 
-                            if isinstance(res, types.CallResult):
-                                reply = message.Yield(msg.request, args=res.results, kwargs=res.kwresults)
-                            else:
-                                reply = message.Yield(msg.request, args=[res])
+                                if msg.receive_progress:
 
-                            try:
-                                self._transport.send(reply)
-                            except SerializationError as e:
-                                # the application-level payload returned from the invoked procedure can't be serialized
-                                reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
-                                                      args=[u'success return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
-                                self._transport.send(reply)
+                                    def progress(*args, **kwargs):
+                                        encrypted_payload = None
+                                        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+                                            if not self._keyring:
+                                                raise Exception(u"trying to send encrypted payload, but no keyring active")
+                                            encrypted_payload = self._keyring.encrypt(False, proc, args, kwargs)
 
-                        def error(err):
-                            errmsg = txaio.failure_message(err)
-                            try:
-                                self.onUserError(err, errmsg)
-                            except:
-                                pass
-                            formatted_tb = None
-                            if self.traceback_app:
-                                formatted_tb = txaio.failure_format_traceback(err)
+                                        if encrypted_payload:
+                                            progress_msg = message.Yield(msg.request,
+                                                                         payload=encrypted_payload.payload,
+                                                                         progress=True,
+                                                                         enc_algo=encrypted_payload.algo,
+                                                                         enc_key=encrypted_payload.pkey,
+                                                                         enc_serializer=encrypted_payload.serializer)
+                                        else:
+                                            progress_msg = message.Yield(msg.request,
+                                                                         args=args,
+                                                                         kwargs=kwargs,
+                                                                         progress=True)
 
-                            del self._invocations[msg.request]
+                                        self._transport.send(progress_msg)
+                                else:
+                                    progress = None
 
-                            reply = self._message_from_exception(
-                                message.Invocation.MESSAGE_TYPE,
-                                msg.request,
-                                err.value,
-                                formatted_tb,
-                            )
+                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(progress, caller=msg.caller, procedure=proc, enc_algo=msg.enc_algo)
 
-                            try:
-                                self._transport.send(reply)
-                            except SerializationError as e:
-                                # the application-level payload returned from the invoked procedure can't be serialized
-                                reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
-                                                      args=[u'error return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
-                                self._transport.send(reply)
-                            # we have handled the error, so we eat it
-                            return None
+                            on_reply = txaio.as_future(endpoint.fn, *invoke_args, **invoke_kwargs)
 
-                        self._invocations[msg.request] = InvocationRequest(msg.request, on_reply)
+                            def success(res):
+                                del self._invocations[msg.request]
 
-                        txaio.add_callbacks(on_reply, success, error)
+                                encrypted_payload = None
+                                if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+                                    if not self._keyring:
+                                        log_msg = u"trying to send encrypted payload, but no keyring active"
+                                        self.log.warn(log_msg)
+                                    else:
+                                        try:
+                                            if isinstance(res, types.CallResult):
+                                                encrypted_payload = self._keyring.encrypt(False, proc, res.results, res.kwresults)
+                                            else:
+                                                encrypted_payload = self._keyring.encrypt(False, proc, [res])
+                                        except Exception as e:
+                                            log_msg = u"failed to encrypt application payload: {}".format(e)
+                                            self.log.warn(log_msg)
+
+                                if encrypted_payload:
+                                    reply = message.Yield(msg.request,
+                                                          payload=encrypted_payload.payload,
+                                                          enc_algo=encrypted_payload.algo,
+                                                          enc_key=encrypted_payload.pkey,
+                                                          enc_serializer=encrypted_payload.serializer)
+                                else:
+                                    if isinstance(res, types.CallResult):
+                                        reply = message.Yield(msg.request,
+                                                              args=res.results,
+                                                              kwargs=res.kwresults)
+                                    else:
+                                        reply = message.Yield(msg.request,
+                                                              args=[res])
+
+                                try:
+                                    self._transport.send(reply)
+                                except SerializationError as e:
+                                    # the application-level payload returned from the invoked procedure can't be serialized
+                                    reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
+                                                          args=[u'success return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
+                                    self._transport.send(reply)
+
+                            def error(err):
+                                del self._invocations[msg.request]
+
+                                errmsg = txaio.failure_message(err)
+
+                                try:
+                                    self.onUserError(err, errmsg)
+                                except:
+                                    pass
+
+                                formatted_tb = None
+                                if self.traceback_app:
+                                    formatted_tb = txaio.failure_format_traceback(err)
+
+                                reply = self._message_from_exception(
+                                    message.Invocation.MESSAGE_TYPE,
+                                    msg.request,
+                                    err.value,
+                                    formatted_tb,
+                                    msg.enc_algo
+                                )
+
+                                try:
+                                    self._transport.send(reply)
+                                except SerializationError as e:
+                                    # the application-level payload returned from the invoked procedure can't be serialized
+                                    reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
+                                                          args=[u'error return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
+                                    self._transport.send(reply)
+                                # we have handled the error, so we eat it
+                                return None
+
+                            self._invocations[msg.request] = InvocationRequest(msg.request, on_reply)
+
+                            txaio.add_callbacks(on_reply, success, error)
 
             elif isinstance(msg, message.Interrupt):
 
@@ -784,7 +977,7 @@ class ApplicationSession(BaseSession):
         Implements :func:`autobahn.wamp.interfaces.ISession.leave`
         """
         if not self._session_id:
-            raise Exception("not joined")
+            raise SessionNotReady(u"session hasn't joined a realm")
 
         if not self._goodbye_sent:
             if not reason:
@@ -796,7 +989,7 @@ class ApplicationSession(BaseSession):
             is_closed = self._transport is None or self._transport.is_closed
             return is_closed
         else:
-            raise SessionNotReady(u"Already requested to close the session")
+            raise SessionNotReady(u"session was alread requested to leave")
 
     def onDisconnect(self):
         """
@@ -815,19 +1008,49 @@ class ApplicationSession(BaseSession):
         if not self._transport:
             raise exception.TransportLost()
 
+        options = kwargs.pop('options', None)
+        if options and not isinstance(options, types.PublishOptions):
+            raise Exception("options must be of type a.w.t.PublishOptions")
+
         request_id = self._request_id_gen.next()
 
-        if 'options' in kwargs and isinstance(kwargs['options'], types.PublishOptions):
-            options = kwargs.pop('options')
-            msg = message.Publish(request_id, topic, args=args, kwargs=kwargs, **options.message_attr())
+        encrypted_payload = None
+        if self._keyring:
+            encrypted_payload = self._keyring.encrypt(True, topic, args, kwargs)
+
+        if encrypted_payload:
+            if options:
+                msg = message.Publish(request_id,
+                                      topic,
+                                      payload=encrypted_payload.payload,
+                                      enc_algo=encrypted_payload.algo,
+                                      enc_key=encrypted_payload.pkey,
+                                      enc_serializer=encrypted_payload.serializer,
+                                      **options.message_attr())
+            else:
+                msg = message.Publish(request_id,
+                                      topic,
+                                      payload=encrypted_payload.payload,
+                                      enc_algo=encrypted_payload.algo,
+                                      enc_key=encrypted_payload.pkey,
+                                      enc_serializer=encrypted_payload.serializer)
         else:
-            options = None
-            msg = message.Publish(request_id, topic, args=args, kwargs=kwargs)
+            if options:
+                msg = message.Publish(request_id,
+                                      topic,
+                                      args=args,
+                                      kwargs=kwargs,
+                                      **options.message_attr())
+            else:
+                msg = message.Publish(request_id,
+                                      topic,
+                                      args=args,
+                                      kwargs=kwargs)
 
         if options and options.acknowledge:
             # only acknowledged publications expect a reply ..
             on_reply = txaio.create_future()
-            self._publish_reqs[request_id] = PublishRequest(request_id, on_reply)
+            self._publish_reqs[request_id] = PublishRequest(request_id, on_reply, was_encrypted=(encrypted_payload is not None))
         else:
             on_reply = None
 
@@ -940,30 +1163,60 @@ class ApplicationSession(BaseSession):
         if not self._transport:
             raise exception.TransportLost()
 
+        options = kwargs.pop('options', None)
+        if options and not isinstance(options, types.CallOptions):
+            raise Exception("options must be of type a.w.t.CallOptions")
+
         request_id = self._request_id_gen.next()
 
-        if 'options' in kwargs and isinstance(kwargs['options'], types.CallOptions):
-            options = kwargs.pop('options')
-            msg = message.Call(request_id, procedure, args=args, kwargs=kwargs, **options.message_attr())
-        else:
-            options = None
-            msg = message.Call(request_id, procedure, args=args, kwargs=kwargs)
+        encrypted_payload = None
+        if self._keyring:
+            encrypted_payload = self._keyring.encrypt(True, procedure, args, kwargs)
 
-        # FIXME
+        if encrypted_payload:
+            if options:
+                msg = message.Call(request_id,
+                                   procedure,
+                                   payload=encrypted_payload.payload,
+                                   enc_algo=encrypted_payload.algo,
+                                   enc_key=encrypted_payload.pkey,
+                                   enc_serializer=encrypted_payload.serializer,
+                                   **options.message_attr())
+            else:
+                msg = message.Call(request_id,
+                                   procedure,
+                                   payload=encrypted_payload.payload,
+                                   enc_algo=encrypted_payload.algo,
+                                   enc_key=encrypted_payload.pkey,
+                                   enc_serializer=encrypted_payload.serializer)
+        else:
+            if options:
+                msg = message.Call(request_id,
+                                   procedure,
+                                   args=args,
+                                   kwargs=kwargs,
+                                   **options.message_attr())
+            else:
+                msg = message.Call(request_id,
+                                   procedure,
+                                   args=args,
+                                   kwargs=kwargs)
+
+        # FIXME: implement call canceling
         # def canceller(_d):
         #   cancel_msg = message.Cancel(request)
         #   self._transport.send(cancel_msg)
         # d = Deferred(canceller)
 
         on_reply = txaio.create_future()
-        self._call_reqs[request_id] = CallRequest(request_id, on_reply, options)
+        self._call_reqs[request_id] = CallRequest(request_id, procedure, on_reply, options)
 
         try:
             # Notes:
             #
             # * this might raise autobahn.wamp.exception.SerializationError
             #   when the user payload cannot be serialized
-            # * we have to setup a PublishRequest() in _publish_reqs _before_
+            # * we have to setup a CallRequest() in _call_reqs _before_
             #   calling transpor.send(), because a mock- or side-by-side transport
             #   will immediately lead on an incoming WAMP message in onMessage()
             #
